-- helper: choose a key from a weights table
local function WeightedPick(weights)
	local candidates = {}
	local total = 0

	for key, weight in pairs(weights) do
		if weight > 0 then
			table.insert(candidates, {key = key, weight = weight})
			total = $ + weight
		end
	end

	if total == 0 then return nil end -- nothing to choose from

	-- Try each entry as a probabilistic choice
	for i = 1, #candidates do
		local chance = FixedDiv(candidates[i].weight * FRACUNIT, total * FRACUNIT)
		if P_RandomChance(chance) then
			return candidates[i].key
		end
	end

	-- Fallback: pick a random key (if all chances fail)
	local fallback = P_RandomKey(#candidates)
	return candidates[fallback + 1].key
end


-- pull pickupgift values from HL_WpnStats for convenience
local PG = {
  ["9mmhandgun"] = HL_WpnStats["9mmhandgun"].primary.pickupgift,
  ["357"]        = HL_WpnStats["357"].primary.pickupgift,
  ["mp5"]        = {
    primary   = HL_WpnStats["mp5"].primary.pickupgift,
    secondary = HL_WpnStats["mp5"].secondary.pickupgift,
  },
  shotgun       = HL_WpnStats["shotgun"].primary.pickupgift,
  crossbow      = HL_WpnStats["crossbow"].primary.pickupgift,
  handgrenade   = HL_WpnStats["handgrenade"].primary.pickupgift,
  satchel       = HL_WpnStats["satchel"].primary.pickupgift,
}

-- 1) Weapon panels
local PanelPickupStats = {
  [MT_AUTOPICKUP] = {
    weights = { mp5=70, shotgun=15, ["9mmhandgun"]=15 },
    defs = {
      mp5 = {
        weapon = "mp5"
      },
      shotgun = {
        weapon = "shotgun"
      },
      ["9mmhandgun"] = {
        weapon = "9mmhandgun"
      },
    }
  },

  [MT_SCATTERPICKUP] = {
    weights = { shotgun=100 },
    defs = {
      shotgun = {
        weapon = "shotgun"
      },
    }
  },

  [MT_BOUNCEPICKUP] = {
    weights = { handgrenade=60, satchel=40 },
    defs = {
      handgrenade = {
        ammo = {
          type={"grenade"},
          give={PG["handgrenade"]},
        }
      },
      satchel = {
        ammo = {
          type={"satchel"},
          give={PG["satchel"]},
        }
      },
    }
  },

  [MT_GRENADEPICKUP] = {
    weights = { handgrenade=60, satchel=40 },
    defs = {
      handgrenade = {
        ammo = {
          type={"grenade"},
          give={PG["handgrenade"]},
        }
      },
      satchel = {
        ammo = {
          type={"satchel"},
          give={PG["satchel"]},
        }
      },
    }
  },

  [MT_EXPLODEPICKUP] = {
    weights = { satchel=70, handgrenade=30 },
    defs = {
      satchel = {
        ammo = {
          type={"satchel"},
          give={PG["satchel"]},
        }
      },
      handgrenade = {
        ammo = {
          type={"grenade"},
          give={PG["handgrenade"]},
        }
      },
    }
  },

  [MT_RAILPICKUP] = {
    weights = { ["357"]=50, crossbow=50 },
    defs = {
      ["357"] = {
        weapon = "357"
      },
      crossbow = {
        weapon = "crossbow"
      },
    }
  },
}

-- weapon‐panel touch handler
local function PanelTouch(special, toucher)
	if not special.valid or not toucher.valid then return end
	local player = toucher.player

	local entry = PanelPickupStats[special.type]
	if not entry then return end

	local choice = WeightedPick(entry.weights)
	local stats  = entry.defs[choice]
	if ApplyPickupStats(player, stats) then
		P_KillMobj(special)
	end

	return true
end

-- install panel hooks
addHook("TouchSpecial", PanelTouch, MT_AUTOPICKUP)
addHook("TouchSpecial", PanelTouch, MT_SCATTERPICKUP)
addHook("TouchSpecial", PanelTouch, MT_BOUNCEPICKUP)
addHook("TouchSpecial", PanelTouch, MT_GRENADEPICKUP)
addHook("TouchSpecial", PanelTouch, MT_EXPLODEPICKUP)
addHook("TouchSpecial", PanelTouch, MT_RAILPICKUP)


-- 2) Ammo rings
local AmmoPickupStats = {
  [MT_INFINITYRING] = {
    weights={ ["9mmhandgun"]=20, ["357"]=15, mp5=15, shotgun=15, crossbow=15, handgrenade=10, satchel=10 },
    defs = {
      ["9mmhandgun"] = { ammo={ type={"9mm"}, give={PG["9mmhandgun"]} } },
      ["357"]        = { ammo={ type={"357"}, give={PG["357"]} } },
      mp5            = { ammo={ type={"9mm","argrenade"}, give={PG["mp5"].primary,PG["mp5"].secondary} } },
      shotgun        = { ammo={ type={"shotgun"}, give={PG["shotgun"]} } },
      crossbow       = { ammo={ type={"bolt"}, give={PG["crossbow"]} } },
      handgrenade    = { ammo={ type={"grenade"}, give={PG["handgrenade"]} } },
      satchel        = { ammo={ type={"satchel"}, give={PG["satchel"]} } },
    }
  },
  [MT_AUTOMATICRING] = {
    weights={ mp5=70, ["9mmhandgun"]=30 },
    defs = {
      mp5 = { ammo={ type={"9mm"}, give={PG["mp5"].primary} } },
      ["9mmhandgun"] = { ammo={ type={"9mm"}, give={PG["9mmhandgun"]} } },
    }
  },
  [MT_SCATTERRING] = {
    weights={ shotgun=100 },
    defs = {
      shotgun = { ammo={ type={"buckshot"}, give={PG["shotgun"]} } },
    }
  },
  [MT_BOUNCERING] = {
    weights={ handgrenade=60, satchel=40 },
    defs = {
      handgrenade = { ammo={ type={"grenade"}, give={PG["handgrenade"]} } },
      satchel     = { ammo={ type={"satchel"}, give={PG["satchel"]} } },
    }
  },
  [MT_GRENADERING] = {
    weights={ handgrenade=30, satchel=40, mp5 = 30 },
    defs = {
      handgrenade = { ammo={ type={"grenade"}, give={PG["handgrenade"]} } },
      satchel     = { ammo={ type={"satchel"}, give={PG["satchel"]} } },
      mp5         = { ammo={ type={"argrenade"}, give={PG["mp5"].secondary} } },
    }
  },
  [MT_EXPLOSIONRING] = {
    weights={ satchel=70, handgrenade=15, mp5 = 15 },
    defs = {
      satchel     = { ammo={ type={"satchel"}, give={PG["satchel"]} } },
      handgrenade = { ammo={ type={"grenade"}, give={PG["handgrenade"]} } },
      mp5         = { ammo={ type={"argrenade"}, give={PG["mp5"].secondary} } },
    }
  },
  [MT_RAILRING] = {
    weights={ ["357"]=50, crossbow=50},
    defs = {
      ["357"] = { ammo={ type={"357"}, give={PG["357"]} } },
      crossbow = { ammo={ type={"bolt"}, give={PG["crossbow"]} } },
    }
  },
}

-- ammo‐ring touch handler
local function RingTouch(special, toucher)
	if not special.valid or not toucher.valid then return end
	local player = toucher.player

	local entry = AmmoPickupStats[special.type]
	if not entry then return end

	local choice = WeightedPick(entry.weights)
	local stats  = entry.defs[choice]
	if ApplyPickupStats(player, stats) then
		P_KillMobj(special)
	end
	return true
end

-- install ammo hooks
addHook("TouchSpecial", RingTouch, MT_INFINITYRING)
addHook("TouchSpecial", RingTouch, MT_AUTOMATICRING)
addHook("TouchSpecial", RingTouch, MT_SCATTERRING)
addHook("TouchSpecial", RingTouch, MT_BOUNCERING)
addHook("TouchSpecial", RingTouch, MT_GRENADERING)
addHook("TouchSpecial", RingTouch, MT_EXPLOSIONRING)
addHook("TouchSpecial", RingTouch, MT_RAILRING)